@@ -11,9 +11,12 @@
 #include "python/descriptor.h"
 #include "python/extension_dict.h"
 #include "python/map.h"
+#include "python/protobuf.h"
 #include "python/repeated.h"
+#include "upb/base/string_view.h"
 #include "upb/message/compare.h"
 #include "upb/message/copy.h"
+#include "upb/message/message.h"
 #include "upb/reflection/def.h"
 #include "upb/reflection/message.h"
 #include "upb/text/encode.h"
@@ -43,12 +46,6 @@ typedef struct {
   size_t type_basicsize;       // sizeof(PyHeapTypeObject)
   traverseproc type_traverse;  // PyTypeObject.tp_traverse
   inquiry type_clear;          // PyTypeObject.tp_clear
-
-  // While we can refer to PY_VERSION_HEX in the limited API, this will give us
-  // the version of Python we were compiled against, which may be different
-  // than the version we are dynamically linked against.  Here we want the
-  // version that is actually running in this process.
-  long python_version_hex;  // PY_VERSION_HEX
 } PyUpb_CPythonBits;
 
 // A global containing the values for this process.
@@ -142,7 +139,6 @@ static bool PyUpb_CPythonBits_Init(PyUpb_CPythonBits* bits) {
 
   sys = PyImport_ImportModule("sys");
   hex_version = PyObject_GetAttrString(sys, "hexversion");
-  bits->python_version_hex = PyLong_AsLong(hex_version);
   ret = true;
 
 err:
@@ -177,8 +173,10 @@ err:
 // The parent may also be non-present, in which case a mutation will trigger a
 // chain reaction.
 typedef struct PyUpb_Message {
-  PyObject_HEAD;
+  // clang-format off
+  PyObject_HEAD
   PyObject* arena;
+  // clang-format on
   uintptr_t def;  // Tagged, low bit 1 == upb_FieldDef*, else upb_MessageDef*
   union {
     // when def is msgdef, the data for this msg.
@@ -314,9 +312,14 @@ static bool PyUpb_Message_LookupName(PyUpb_Message* self, PyObject* py_name,
 static bool PyUpb_Message_InitMessageMapEntry(PyObject* dst, PyObject* src) {
   if (!src || !dst) return false;
 
-  PyObject* ok = PyObject_CallMethod(dst, "CopyFrom", "O", src);
-  if (!ok) return false;
-  Py_DECREF(ok);
+  if (PyDict_Check(src)) {
+    bool ok = PyUpb_Message_InitAttributes(dst, NULL, src) >= 0;
+    if (!ok) return false;
+  } else {
+    PyObject* ok = PyObject_CallMethod(dst, "CopyFrom", "O", src);
+    if (!ok) return false;
+    Py_DECREF(ok);
+  }
 
   return true;
 }
@@ -450,13 +453,16 @@ static bool PyUpb_Message_InitMessageAttribute(PyObject* _self, PyObject* name,
     const upb_MessageDef* msgdef = upb_FieldDef_MessageSubDef(field);
     if (upb_MessageDef_WellKnownType(msgdef) == kUpb_WellKnown_Struct) {
       ok = PyObject_CallMethod(submsg, "_internal_assign", "O", value);
-      if (!ok && PyDict_Size(value) == 1 &&
-          PyDict_Contains(value, PyUnicode_FromString("fields"))) {
-        // Fall back to init as normal message field.
-        PyErr_Clear();
-        PyObject* tmp = PyUpb_Message_Clear((PyUpb_Message*)submsg);
-        Py_DECREF(tmp);
-        ok = PyUpb_Message_InitAttributes(submsg, NULL, value) >= 0;
+      if (!ok && PyDict_Size(value) == 1) {
+        PyObject* fields_str = PyUnicode_FromString("fields");
+        if (PyDict_Contains(value, fields_str)) {
+          // Fall back to init as normal message field.
+          PyErr_Clear();
+          PyObject* tmp = PyUpb_Message_Clear((PyUpb_Message*)submsg);
+          Py_DECREF(tmp);
+          ok = PyUpb_Message_InitAttributes(submsg, NULL, value) >= 0;
+        }
+        Py_DECREF(fields_str);
       }
     } else {
       ok = PyUpb_Message_InitAttributes(submsg, NULL, value) >= 0;
@@ -575,9 +581,7 @@ static bool PyUpb_Message_IsEmpty(const upb_Message* msg,
   upb_MessageValue val;
   if (upb_Message_Next(msg, m, ext_pool, &f, &val, &iter)) return false;
 
-  size_t len;
-  (void)upb_Message_GetUnknown(msg, &len);
-  return len == 0;
+  return !upb_Message_HasUnknown(msg);
 }
 
 static bool PyUpb_Message_IsEqual(PyUpb_Message* m1, PyObject* _m2) {
@@ -678,8 +682,10 @@ static void PyUpb_Message_SyncSubobjs(PyUpb_Message* self);
  * the set state (having a non-owning pointer to self->ptr.msg).
  */
 static void PyUpb_Message_Reify(PyUpb_Message* self, const upb_FieldDef* f,
-                                upb_Message* msg) {
+                                upb_Message* msg, PyUpb_WeakMap* subobj_map,
+                                intptr_t iter) {
   assert(f == PyUpb_Message_GetFieldDef(self));
+  PyUpb_WeakMap_DeleteIter(subobj_map, &iter);
   if (!msg) {
     const upb_MessageDef* msgdef = PyUpb_Message_GetMsgdef((PyObject*)self);
     const upb_MiniTable* layout = upb_MessageDef_MiniTable(msgdef);
@@ -731,17 +737,18 @@ static void PyUpb_Message_SyncSubobjs(PyUpb_Message* self) {
     if (upb_FieldDef_HasPresence(f) && !upb_Message_HasFieldByDef(msg, f))
       continue;
     upb_MessageValue msgval = upb_Message_GetFieldByDef(msg, f);
-    PyUpb_WeakMap_DeleteIter(subobj_map, &iter);
     if (upb_FieldDef_IsMap(f)) {
       if (!msgval.map_val) continue;
-      PyUpb_MapContainer_Reify(obj, (upb_Map*)msgval.map_val);
+      PyUpb_MapContainer_Reify(obj, (upb_Map*)msgval.map_val, subobj_map, iter);
     } else if (upb_FieldDef_IsRepeated(f)) {
       if (!msgval.array_val) continue;
-      PyUpb_RepeatedContainer_Reify(obj, (upb_Array*)msgval.array_val);
+      PyUpb_RepeatedContainer_Reify(obj, (upb_Array*)msgval.array_val,
+                                    subobj_map, iter);
     } else {
       PyUpb_Message* sub = (void*)obj;
       assert(self == sub->ptr.parent);
-      PyUpb_Message_Reify(sub, f, (upb_Message*)msgval.msg_val);
+      PyUpb_Message_Reify(sub, f, (upb_Message*)msgval.msg_val, subobj_map,
+                          iter);
     }
   }
 
@@ -826,18 +833,7 @@ static void PyUpb_Message_Dealloc(PyObject* _self) {
   }
 
   Py_DECREF(self->arena);
-
-  // We do not use PyUpb_Dealloc() here because Message is a base type and for
-  // base types there is a bug we have to work around in this case (see below).
-  PyTypeObject* tp = Py_TYPE(self);
-  freefunc tp_free = PyType_GetSlot(tp, Py_tp_free);
-  tp_free(self);
-
-  if (cpython_bits.python_version_hex >= 0x03080000) {
-    // Prior to Python 3.8 there is a bug where deallocating the type here would
-    // lead to a double-decref: https://bugs.python.org/issue37879
-    Py_DECREF(tp);
-  }
+  PyUpb_Dealloc(self);
 }
 
 PyObject* PyUpb_Message_Get(upb_Message* u_msg, const upb_MessageDef* m,
@@ -982,10 +978,12 @@ int PyUpb_Message_SetFieldValue(PyObject* _self, const upb_FieldDef* field,
       if (PyObject_HasAttrString(sub_message, "_internal_assign")) {
         PyObject* ok =
             PyObject_CallMethod(sub_message, "_internal_assign", "O", value);
+        Py_DECREF(sub_message);
         if (!ok) return -1;
         Py_DECREF(ok);
         return 0;
       }
+      Py_DECREF(sub_message);
     }
     PyErr_Format(exc,
                  "Assignment not allowed to message "
@@ -1115,7 +1113,9 @@ static PyObject* PyUpb_Message_Contains(PyObject* _self, PyObject* arg) {
       PyUpb_Message* self = (void*)_self;
       if (PyUpb_Message_IsStub(self)) Py_RETURN_FALSE;
       PyObject* items = PyObject_CallMethod(_self, "items", NULL);
-      return PyBool_FromLong(PySequence_Contains(items, arg));
+      int ret = PySequence_Contains(items, arg);
+      Py_DECREF(items);
+      return PyBool_FromLong(ret);
     }
     default:
       // For other messages, check with HasField.
@@ -1397,18 +1397,17 @@ static PyObject* PyUpb_Message_Clear(PyUpb_Message* self) {
 
     while (PyUpb_WeakMap_Next(subobj_map, &key, &obj, &iter)) {
       const upb_FieldDef* f = key;
-      PyUpb_WeakMap_DeleteIter(subobj_map, &iter);
       if (upb_FieldDef_IsMap(f)) {
         assert(upb_Message_GetFieldByDef(msg, f).map_val == NULL);
-        PyUpb_MapContainer_Reify(obj, NULL);
+        PyUpb_MapContainer_Reify(obj, NULL, subobj_map, iter);
       } else if (upb_FieldDef_IsRepeated(f)) {
         assert(upb_Message_GetFieldByDef(msg, f).array_val == NULL);
-        PyUpb_RepeatedContainer_Reify(obj, NULL);
+        PyUpb_RepeatedContainer_Reify(obj, NULL, subobj_map, iter);
       } else {
         assert(!upb_Message_HasFieldByDef(msg, f));
         PyUpb_Message* sub = (void*)obj;
         assert(self == sub->ptr.parent);
-        PyUpb_Message_Reify(sub, f, NULL);
+        PyUpb_Message_Reify(sub, f, NULL, subobj_map, iter);
       }
     }
   }
@@ -1483,7 +1482,8 @@ static PyObject* PyUpb_Message_DiscardUnknownFields(PyUpb_Message* self,
                                                     PyObject* arg) {
   PyUpb_Message_EnsureReified(self);
   const upb_MessageDef* msgdef = _PyUpb_Message_GetMsgdef(self);
-  upb_Message_DiscardUnknown(self->ptr.msg, msgdef, 64);
+  const upb_DefPool* ext_pool = upb_FileDef_Pool(upb_MessageDef_File(msgdef));
+  upb_Message_DiscardUnknown(self->ptr.msg, msgdef, ext_pool, 64);
   Py_RETURN_NONE;
 }
 
