@@ -19,9 +19,4 @@ fn main() {
         .file("libupb/third_party/utf8_range/utf8_range.c")
         .define("UPB_BUILD_API", Some("1"))
         .compile("libupb");
-
-    let path = std::path::absolute("libupb").expect("Failed to get full path to libupb");
-
-    println!("cargo:include={}", path.display());
-    println!("cargo:version={}", VERSION);
 }
